/**
  *  \file arch/unix/file.cc
  *  \brief File I/O based upon POSIX file descriptors
  *
  *  This provides an implementation of a file I/O class using POSIX
  *  file descriptors. It is some 800 bytes smaller than the C++
  *  iostream-based one, and probably faster (we don't need the iostream
  *  buffering, and we add another buffering layer when we need to).
  */

#include <cerrno>
#include <cstring>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include "cpluslib/string.h"
#include "io/streams.h"
#include "util/except.h"
#include "util/nls.h"
#include "arch/platform.h"

namespace {
    /** \class PosixFileStream
        \brief Stream reading from a file
    
        This is a file stream using POSIX file descriptors. */
    class PosixFileStream : public Stream {
        /** File descriptor. */
        int fd;

        /** File name. UTF-8 encoding. */
        const string_t name;

        /** Mode. */
        int active_mode;

        void init(const char* sysName, int mode);
        std::size_t error();
     public:
        PosixFileStream(string_t name, int mode);
        ~PosixFileStream();
    
        std::size_t read(char* buffer, std::size_t max);
        std::size_t write(const char* buffer, std::size_t max);
        int getCapabilities() const;
        void seek(fpos_t pos);
        fpos_t getPos();
        fpos_t getSize();
        string_t getName();
    };
}

Ptr<Stream>
Stream::openFile(const string_t& name, int mode)
{
    return new PosixFileStream(name, mode);
}

/** Constructor.
    \param name File name in UTF-8 encoding.
    \param mode Open mode */
PosixFileStream::PosixFileStream(const string_t name, const int mode)
    : name(name)
{
    string_t sysName;
    if (!Platform::convertUtf8ToFileName(sysName, name)) {
        throw FileOpenException(name, _("File name not supported"));
    }
    init(sysName.c_str(), mode);
}

PosixFileStream::~PosixFileStream()
{
    if (fd >= 0) {
        ::close(fd);
    }
}

/** Open file. This converts the mode and invokes the actual system call.
    It sets \c fd accordingly.
    \param sysName Name in system encoding
    \param mode    Mode */
void
PosixFileStream::init(const char*const sysName, const int mode)
{
    if (mode & C_CREATE) {
        /* create */
        active_mode = C_READ + C_WRITE + C_SEEK;
        fd = ::creat(sysName, 0666);
    } else {
        /* open existing */
        int open_mode = 0;
        if ((mode & (C_READ + C_WRITE)) == (C_READ + C_WRITE)) {
            active_mode = C_READ + C_WRITE + C_SEEK;
            open_mode = O_RDWR;
        } else if (mode & C_WRITE) {
            active_mode = C_WRITE + C_SEEK;
            open_mode = O_WRONLY;
        } else {
            active_mode = C_READ + C_SEEK;
            open_mode = O_RDONLY;
        }
        fd = ::open(sysName, open_mode);
    }

    if (fd < 0) {
        throw FileOpenException(name, Platform::convertFileNameToUtf8(std::strerror(errno)));
    }
}

std::size_t
PosixFileStream::error()
{
    throw FileIOException(*this, Platform::convertFileNameToUtf8(std::strerror(errno)));
}

std::size_t
PosixFileStream::read(char* buffer, std::size_t max)
{
    ssize_t n = ::read(fd, buffer, max);
    if (n < 0)
        return error();
    else
        return n;
}

std::size_t
PosixFileStream::write(const char* buffer, std::size_t max)
{
    ssize_t n = ::write(fd, buffer, max);
    if (n < 0)
        return error();
    else
        return n;
}

void
PosixFileStream::seek(fpos_t pos)
{
    if (::lseek(fd, pos, SEEK_SET) < 0)
        error();
}

PosixFileStream::fpos_t
PosixFileStream::getPos()
{
    return ::lseek(fd, 0, SEEK_CUR);
}

PosixFileStream::fpos_t
PosixFileStream::getSize()
{
    struct stat st;
    if (::fstat(fd, &st) == 0) {
        return st.st_size;
    } else {
        return -1;
    }
}

int
PosixFileStream::getCapabilities() const
{
    return active_mode;
}

string_t
PosixFileStream::getName()
{
    return name;
}
