/**
  *  \file client/dialogs/combatdlg.cc
  *  \brief Combat Overview Dialog
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/combatdlg.h"
#include "client/widgets/combatdiagram.h"
#include "client/widgets/racelist.h"
#include "game/vcrdb.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/uiglobal.h"
#include "ui/widget-busy.h"
#include "ui/widgets/button.h"
#include "ui/widgets/keytranslator.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/progress.h"
#include "ui/widgets/checkbox.h"
#include "ui/richdocument.h"
#include "game/racename.h"
#include "util/misc.h"

namespace {
    class WCombatDiagramDialog : public UIWindow {
        ObjectHolder h;
        WCombatDiagramWidget* diagWidget;
        GClassicVcrDatabase& db;
        UICheckbox* colorCheckbox;
        int chosenBattle;

        void init(GClassicVcrDatabase& db, WCombatDiagram& diag);
        void onClickBattle();
        void onToggleTeamColors();
        void onScore();
     public:
        WCombatDiagramDialog(GClassicVcrDatabase& db, WCombatDiagram& diag);
        int getChosenBattle() const
            { return chosenBattle; }
    };
}

WCombatDiagramDialog::WCombatDiagramDialog(GClassicVcrDatabase& db, WCombatDiagram& diag)
    : UIWindow(_("Combat Overview"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      db(db),
      chosenBattle(-1)
{
    init(db, diag);
}

void
WCombatDiagramDialog::init(GClassicVcrDatabase& db, WCombatDiagram& diag)
{
    // VBox
    //   WCombatDiagramWidget
    //   WRaceList
    //   HBox
    //     UIButton "OK"
    //     UICheckbox "Team colors"
    //     UISpacer
    setLayout(UIVBoxLayout::instance5);

    diagWidget = &h.add(new WCombatDiagramWidget(db, diag));
    diagWidget->sig_click_battle.add(this, &WCombatDiagramDialog::onClickBattle);
    add(*diagWidget);
    add(h.add(new WRaceList(30*font_ems[FONT_NORMAL], WRaceList::FlowLayout | WRaceList::ShowRaceNames, 1, NUM_OWNERS)));

    colorCheckbox = &h.add(new UICheckbox(0, _("Team colors"), 't', 2));
    colorCheckbox->setValue(diagWidget->getTeamColorState());
    colorCheckbox->sig_changed.add(this, &WCombatDiagramDialog::onToggleTeamColors);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UIButton(2, _("Scores"), 's')).addCall(this, &WCombatDiagramDialog::onScore));
    g.add(*colorCheckbox);
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("Close"), SDLK_RETURN)).addStop());
    add(g);
    add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));

    GfxRect pos = ui_root->getExtent();
    pos.grow(-10, -10);
    setExtent(pos);
}

void
WCombatDiagramDialog::onClickBattle()
{
    // Must save the chosen battle because closing the window will
    // invalidate it inside the diagWidget.
    chosenBattle = diagWidget->getHoverSlot();
    stop(2);
}

void
WCombatDiagramDialog::onToggleTeamColors()
{
    diagWidget->setTeamColorState(colorCheckbox->getValue());
}

void
WCombatDiagramDialog::onScore()
{
    // Compute scores
    int32_t scores[NUM_PLAYERS][GVcrEntry::NumScores];
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        for (int j = 0; j < GVcrEntry::NumScores; ++j) {
            scores[i][j] = 0;
        }
    }
    GPlayerSet players;
    for (int i = 0, n = db.getCount(); i < n; ++i) {
        const GVcrEntry& e = db.getEntry(i);
        if (e.getStatus() != 0 && e.getStatus() != VcrPlayer::VCRS_INVALID) {
            for (int side = 0; side < 2; ++side) {
                int player = e.getVcr().objects[side].owner;
                if (player > 0 && player <= NUM_PLAYERS) {
                    e.computeScores(scores[player-1], side);
                    players += player;
                }
            }
        }
    }

    // Find columns. Also convert millipoints to points.
    enum {
        ExperienceColumn      = 1,
        BuildPointColumn      = 2,
        BuildPointRangeColumn = 4,
        TonsColumn            = 8
    };
    int columns = 0;
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        scores[i][GVcrEntry::EarnedBuildMillipointsMin] /= 1000;
        scores[i][GVcrEntry::EarnedBuildMillipointsMax] /= 1000;
        if (scores[i][GVcrEntry::EarnedBuildMillipointsMin] > 0) {
            if (scores[i][GVcrEntry::EarnedBuildMillipointsMax] != scores[i][GVcrEntry::EarnedBuildMillipointsMin]) {
                columns |= BuildPointRangeColumn;
            } else {
                columns |= BuildPointColumn;
            }
        }
        if (scores[i][GVcrEntry::EarnedExperience] > 0) {
            columns |= ExperienceColumn;
        }
        if (scores[i][GVcrEntry::EarnedTonsDestroyed] > 0) {
            columns |= TonsColumn;
        }
    }
    if (columns == 0) {
        // This should never happen as we always compute tons. Therefore,
        // no need to do any fancy error messaging.
        return;
    }

    // Find column widths
    int nameWidth = 0;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        nameWidth = std::max(fonts[FONT_NORMAL]->getTextWidth(player_racenames.getFullName(i)), nameWidth);
    }
    const char* expTitle = _("Experience");
    int expWidth = (columns & ExperienceColumn) != 0 ? std::max(fonts[FONT_NORMAL_BOLD]->getTextWidth(expTitle), font_ems[FONT_NORMAL]*5) + 5 : 0;
    const char* bpTitle = _("Build Points");
    int bpWidth = (columns & (BuildPointRangeColumn | BuildPointColumn)) != 0
        ? std::max(fonts[FONT_NORMAL_BOLD]->getTextWidth(bpTitle),
                   font_ems[FONT_NORMAL] * ((columns & BuildPointRangeColumn) != 0 ? 10 : 5)) + 5
        : 0;
    const char* tonsTitle = _("Tons sunk");
    int tonsWidth = (columns & TonsColumn) != 0 ? std::max(fonts[FONT_NORMAL_BOLD]->getTextWidth(tonsTitle), font_ems[FONT_NORMAL]*5) + 5 : 0;

    int width1 = nameWidth + expWidth;
    int width2 = width1 + bpWidth;
    int totalWidth = width2 + tonsWidth;

    // Generate document
    UIRichDocument d(GfxPoint(totalWidth, 10), 0);
    RichDocument& doc(d.getDocument());
    doc.setPageWidth(totalWidth);
    if (columns & ExperienceColumn) {
        doc.addCentered(nameWidth + expWidth/2,
                        RichText(expTitle).withAttribute(RichTextStyleAttribute::Bold).withAttribute(RichTextStyleAttribute::Underline));
    }
    if (columns & (BuildPointRangeColumn | BuildPointColumn)) {
        doc.addCentered(width1 + bpWidth/2,
                        RichText(bpTitle).withAttribute(RichTextStyleAttribute::Bold).withAttribute(RichTextStyleAttribute::Underline));
    }
    if (columns & TonsColumn) {
        doc.addCentered(width2 + tonsWidth/2,
                        RichText(tonsTitle).withAttribute(RichTextStyleAttribute::Bold).withAttribute(RichTextStyleAttribute::Underline));
    }
    doc.addNewline();

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (players.contains(i)) {
            int32_t* ps = scores[i-1];
            doc.add(player_racenames.getFullName(i));
            if (columns & ExperienceColumn) {
                doc.addRight(width1-5, numToString(ps[GVcrEntry::EarnedExperience]));
            }
            if (columns & BuildPointRangeColumn) {
                doc.addRight(width1 + bpWidth/2, string_t(numToString(ps[GVcrEntry::EarnedBuildMillipointsMin]) + ".."));
                doc.add(numToString(ps[GVcrEntry::EarnedBuildMillipointsMax]));
            } else if (columns & BuildPointColumn) {
                doc.addRight(width2-5, numToString(ps[GVcrEntry::EarnedBuildMillipointsMin]));
            }
            if (columns & TonsColumn) {
                doc.addRight(totalWidth-5, numToString(ps[GVcrEntry::EarnedTonsDestroyed]));
            }
            doc.addNewline();
        }
    }
    d.adjustToDocumentSize();

    // Display it
    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Combat Score Overview"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);
    win.add(d);

    UIGroup& group = h.add(new UIGroup(UIHBoxLayout::instance5));
    group.add(h.add(new UISpacer()));
    group.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    group.add(h.add(new UISpacer()));
    win.add(group);
    win.add(h.add(new UIKeyTranslator())
            .add(' ', 1, true)
            .add(SDLK_ESCAPE, 1, true));

    win.pack();
    ui_root->centerWidget(win);
    ui_root->runModal(win);
}

/** Combat overview dialog. Displays a WCombatDiagramWidget.
    \param db VCRs to display
    \return clicked VCR number [0,db.getCount()), if any; otherwise -1 */
int
doCombatDialog(GClassicVcrDatabase& db)
{
    // Need at least two VCRs
    if (db.getCount() < 2)
        return -1;

    // Initialize
    WCombatDiagram diag;
    {
        ProgressMonitor progress;
        UIBusyWidget busy;
        ui_root->add(busy);
        ui_root->performDeferredRedraws();
        progress.sig_tick.add(&busy, &UIBusyWidget::tick);

        diag.init(db, progress);
    }

    // Display dialog
    WCombatDiagramDialog win(db, diag);
    if (ui_root->runModal(win) == 2) {
        return win.getChosenBattle();
    } else {
        return -1;
    }
}
