/**
  *  \file client/lrupredef.cc
  *  \brief Least-Recently-Used Lists - Predefined Lists
  *
  *  This module defines the global LRUList instances used in the PCC2
  *  GUI. It also provides loading of predefined item lists from the
  *  "expr.cc"/"expr.usr" file through a modinit hook.
  */

#include "client/lru.h"
#include "cpluslib/format.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/filepars.h"
#include "util/modinit.h"
#include "util/nls.h"

/**************************** Predefined Lists ***************************/

LRUList predef_shiplabels, predef_planetlabels, predef_search;
LRUList lru_shiplabels, lru_planetlabels, lru_search;

static LRUList*
getListByName(const string_t& str)
{
    if (str == "SHIPLABELS") {
        return &predef_shiplabels;
    }
    if (str == "PLANETLABELS") {
        return &predef_planetlabels;
    }
    if (str == "FIND") {
        return &predef_search;
    }
    return 0;
}

namespace {
    class ListFileParser : public FileParser {
        LRUList* section;
     public:
        ListFileParser();
        void syntaxError(int line_id);
        void process(int line_id, string_t line);
    };
}

ListFileParser::ListFileParser()
    : FileParser(";#"),
      section(0)
{ }

void
ListFileParser::syntaxError(int line_id)
{
    console.write(LOG_WARN, format(_("%s:%d: file format error -- line ignored"), getFileName(), line_id));
}

void
ListFileParser::process(int line_id, string_t line)
{
    string_t::size_type n = line.find_first_not_of(" \t");
    if (n == string_t::npos) {
        // cannot happen
    } else if (line[n] == '[') {
        // section
        string_t::size_type p = line.find(']', n);
        if (p == string_t::npos) {
            syntaxError(line_id);
        } else {
            section = getListByName(strUCase(string_t(line, n+1, p-n-1)));
        }
    } else if (section != 0) {
        // line in section
        string_t::size_type p = line.find("  ", n);
        if (p == string_t::npos) {
            syntaxError(line_id);
        } else {
            section->pushNew(new LRUListItem(string_t(line, n, p-n), strTrim(string_t(line, p+1))));
        }
    } else {
        // unknown section
    }
}

/***************************** Initialisation ****************************/

static void
loadLRULists()
{
    predef_shiplabels.clear();
    predef_planetlabels.clear();
    predef_search.clear();

    ListFileParser().parseGameFile("expr.cc");
    ListFileParser().parseGameFile("expr.usr");
}

const ModuleInitializer mod_lru[] = {
    { OnEnterNoGame | OnEnterDirectory, loadLRULists },
    { 0, 0 },
};
