/**
  *  \file game/obj.h
  *  \brief Game Object Base Classes
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJ_H
#define PCC_V2_GAME_OBJ_H

#include "game/coord.h"
#include "cpluslib/string.h"
#include "cpluslib/object.h"
#include "cpluslib/types.h"
#include "cpluslib/sigslot.h"

/// Game object.
class GObject : public Object {
    bool is_played;
    bool is_marked;
    bool is_dirty;

    GObject(const GObject&);
    GObject& operator=(const GObject&);
 public:
    /** Constructor. */
    GObject()
        : is_played(false), is_marked(false), is_dirty(false)
        { }

    /** Destructor. */
    virtual ~GObject();

    // Object:
    /** Convert to string. */
    string_t toString();

    // GObject interface:
    /** Get name of this object.
        \return name, e.g. "USS Endeavour" */
    virtual string_t getName() const = 0;

    /** Get full name of this object.
        \return full name, e.g. "Ship #42: USS Endeavour (Ruby, Tholian)" */
    virtual string_t getFullName() const = 0;

    /** Get Id number of this object. */
    virtual int getId() const = 0;

    /** Get owner of this object. */
    virtual int getOwner() const = 0;

    /** Check whether this object exists.
        \return true if this object exists, false if not. */
    virtual bool isExistant() const = 0;


    /** Check whether this object is being played. */
    bool isPlayed() const
        { return is_played; }

    /** Set whether this object is being played. */
    void setIsPlayed(bool n)
        { is_played = n; }

    
    /** Mark object clean. */
    void markClean()
        { is_dirty = false; }

    /** Mark object dirty. */
    void markDirty()
        { is_dirty = true; }

    /** Check whether specific area of object is dirty. */
    bool isDirty() const
        { return is_dirty; }

    
    /** Check whether object is marked. */
    bool isMarked() const
        { return is_marked; }

    /** Set selection status. */
    void setIsMarked(bool n)
        {
            markDirty();
            is_marked = n;
        }

    void notifyListeners();

    /** Signal for object changes.
        \param int getId() of the object */
    Signal1<int> sig_changed;
};


/// Physical game object on map.
class GMapObject : public GObject {
 public:
    /** Destructor. */
    virtual ~GMapObject();

    /** Get position in game universe. */
    virtual GPoint getPos() const = 0;
};


/// Circular game object on map.
class GCircularObject : public GMapObject {
 public:
    virtual ~GCircularObject();

    /** Get object radius in light-years. */
    virtual int getRadius() const = 0;

    /** Get squared radius in light-years. This is used for objects that have fractional radius,
        such as minefields. */
    virtual int32_t getRadiusSquared() const = 0;
};

#endif
