/**
  *  \file int/csource.cc
  *  \brief Interpreter: Command Sources
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/csource.h"
#include "io/textfile.h"

/**************************** IntCommandSource ***************************/

/** Default constructor. */
IntCommandSource::IntCommandSource()
    : tok(string_t()),
      line_nr(0),
      eof(true)
{ }

/** Destructor. */
IntCommandSource::~IntCommandSource()
{ }

/** Set next line. Call this from readNextLine. */
void
IntCommandSource::setNextLine(string_t s)
{
    ++line_nr;
    eof = false;
    tok = IntTokenizer(s);
}


/************************* IntMemoryCommandSource ************************/

/** Constructor. Makes a blank command source. */
IntMemoryCommandSource::IntMemoryCommandSource()
    : index(0)
{ }

/** Constructor. Makes a command source containing a single line.
    \param line The line */
IntMemoryCommandSource::IntMemoryCommandSource(string_t line)
    : lines(1, line),
      index(0)
{ }

/** Add line to this command source.
    \param line The line */
void
IntMemoryCommandSource::addLine(const string_t& line)
{
    lines.push_back(line);
}

void
IntMemoryCommandSource::readNextLine()
{
    if (index < lines.size()) {
        setNextLine(lines[index++]);
    } else {
        setEOF();
    }
}

bool
IntMemoryCommandSource::setEncoding(bool /*utf8*/, CharacterSet /*cs*/)
{
    return false;
}


/************************** IntFileCommandSource *************************/

/** File command source.
    \param tf Text file to read from */
IntFileCommandSource::IntFileCommandSource(TextFile& tf)
    : tf(tf)
{ }

void
IntFileCommandSource::readNextLine()
{
    string_t line;
    if (tf.getLine(line))
        setNextLine(line);
    else
        setEOF();
}

bool
IntFileCommandSource::setEncoding(bool utf8, CharacterSet cs)
{
    if (utf8)
        tf.setUtf8();
    else
        tf.setCharacterSet(cs);
    return true;
}
