/**
  *  \file int/unary.cc
  *  \brief Interpreter: Unary Operations
  *
  *  (c) 2001-2012 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "int/unary.h"
#include "int/value.h"
#include "cpluslib/misc.h"
#include "game/formula.h"       // PI
#include "int/interror.h"
#include "cpluslib/string.h"
#include "util/console.h"
#include "util/atom.h"
#include "int/keymap.h"
#include "int/compound.h"
#include "cpluslib/utf8.h"
#include "int/array.h"

/********************************* Naming ********************************/

static const char* const unary_names[] = {
    "not",
    "bool",
    "neg",
    "pos",
    "sin",
    "cos",
    "tan",
    "zap",
    "abs",
    "exp",
    "log",
    "bitnot",
    "isempty",
    "isnum",
    "isstr",
    "asc",
    "chr",
    "str",
    "sqrt",
    "trunc",
    "round",
    "ltrim",
    "rtrim",
    "lrtrim",
    "length",
    "val",
    "trace",
    "not2",
    "atom",
    "atomstr",
    "keycreate",
    "keylookup",
    "inc",
    "dec",
    "isproc",
    "filenr",
    "isarray",
};

/** Get name for an unary operation. This is used for disassembling.
    \param op Operation, IntUnaryOperation */
const char*
getUnaryName(uint8_t op)
{
    if (op < countof(unary_names))
        return unary_names[op];
    else
        return "?";
}

/******************************* Execution *******************************/

namespace {
    /** Arithmetic status. */
    enum Arithmetic {
        ariBad,                 ///< Bad argument combination.
        ariNull,                ///< Either argument is null.
        ariInt,                 ///< Use integer arithmetic.
        ariFloat                ///< Use float arithmetic.
    };

    /** Arithmetic argument. */
    struct ArithmeticArg {
        int ia;
        double fa;
    };
}

/** Classify argument for arithmetic operations.
    \param arg [out] Argument for arithmetic
    \param a [in] User-supplied argument
    \return status */
static Arithmetic
checkArithmetic(ArithmeticArg& arg, IntValue* a)
{
    if (a == 0) {
        return ariNull;
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(a)) {
        arg.ia = iv->getValue();
        return ariInt;
    } else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(a)) {
        arg.fa = fv->getValue();
        return ariFloat;
    } else {
        return ariBad;
    }
}

/** Prepare for trigonometry. Checks whether the argument is of the correct
    type, has reasonable range, and converts it from degrees to radians.
    \param arg user-supplied argument
    \return argument to pass to C++ math library
    \throws IntError if bad user-supplied argument */
static double
prepareTrig(IntValue* arg)
{
    // Fetch value
    double value;
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(arg))
        value = iv->getValue();
    else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(arg))
        value = fv->getValue();
    else
        throw IntError::typeError(IntError::tyNumeric);

    // Range check
    if (std::fabs(value) > 1.0E+6)
        throw IntError::rangeError();

    return value * (PI / 180.0);
}

static IntValue*
FNot(IntValue* arg)
{
    // logical not
    int value = getBoolValue(arg);
    if (value >= 0)
        value ^= 1;
    return makeBoolValue(value);
}

static IntValue*
FBool(IntValue* arg)
{
    // conversion to bool, i.e. double not
    return makeBoolValue(getBoolValue(arg));
}

static IntValue*
FNeg(IntValue* arg)
{
    // arithmetic negation
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(-a.ia);
     case ariFloat:
        return makeFloatValue(-a.fa);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FPos(IntValue* arg)
{
    // arithmetic "identity", i.e. just check type
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(+a.ia);
     case ariFloat:
        return makeFloatValue(+a.fa);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FSin(IntValue* arg)
{
    // Sine
    if (!arg)
        return 0;
    else
        return makeFloatValue(std::sin(prepareTrig(arg)));
}

static IntValue*
FCos(IntValue* arg)
{
    // Cosine
    if (!arg)
        return 0;
    else
        return makeFloatValue(std::cos(prepareTrig(arg)));
}

static IntValue*
FTan(IntValue* arg)
{
    // Tangent
    if (!arg) {
        return 0;
    } else {
        double a = prepareTrig(arg);
        double s = std::sin(a);
        double c = std::cos(a);
        if (std::fabs(c) < 1.0E-06)          // FIXME?, see FDivide
            throw IntError("Divide by zero");
        else
            return makeFloatValue(s/c);
    }
}

static IntValue*
FZap(IntValue* arg)
{
    // Turn false into null
    if (!arg)
        return 0;
    else if (getBoolValue(arg) == 0)
        return 0;
    else
        return arg->clone();
}

static IntValue*
FAbs(IntValue* arg)
{
    // Absolute value
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(std::abs(a.ia));
     case ariFloat:
        return makeFloatValue(std::fabs(a.fa));
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FExp(IntValue* arg)
{
    // Exponential
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeFloatValue(std::exp(double(a.ia)));
     case ariFloat:
        return makeFloatValue(std::exp(a.fa));
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FLog(IntValue* arg)
{
    // Logarithm
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        if (a.ia <= 0)
            throw IntError::rangeError();
        else
            return makeFloatValue(std::log(double(a.ia)));
     case ariFloat:
        if (a.fa <= 0)
            throw IntError::rangeError();
        else
            return makeFloatValue(std::log(a.fa));
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FBitNot(IntValue* arg)
{
    // Bitwise negation
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(~a.ia);
     case ariFloat:
     default:
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FIsEmtpy(IntValue* arg)
{
    // Check emptiness (null)
    return makeBoolValue(arg == 0);
}

static IntValue*
FIsNum(IntValue* arg)
{
    // Check numericness
    // FIXME: PCC 1.x returns False for bools
    return makeBoolValue(arg != 0
                         && (dynamic_cast<IntIntValue*>(arg) != 0
                             || dynamic_cast<IntFloatValue*>(arg) != 0));
}

static IntValue*
FIsString(IntValue* arg)
{
    // Check string
    return makeBoolValue(arg != 0
                         && dynamic_cast<IntStringValue*>(arg) != 0);
}

static IntValue*
FAsc(IntValue* arg)
{
    // Get ASCII/Unicode code of (stringified) arg
    if (arg == 0)
        return 0;

    string_t s = arg->toString(false);
    Utf8Reader rdr(s);
    if (rdr.hasMore())
        return makeIntValue(rdr.get());
    else
        return 0;
}

static IntValue*
FChr(IntValue* arg)
{
    // Get character from ASCII/Unicode code
    if (arg == 0) {
        return 0;
    } else if (IntIntValue* iv = dynamic_cast<IntIntValue*>(arg)) {
        string_t tmp;
        appendUtf8(tmp, iv->getValue());
        return makeStringValue(tmp);
    } else {
        throw IntError::typeError(IntError::tyInteger);
    }
}

static IntValue*
FStr(IntValue* arg)
{
    // Simple stringification
    if (arg == 0)
        return 0;
    return makeStringValue(arg->toString(false));
}

static IntValue*
FSqrt(IntValue* arg)
{
    // Square root
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        if (a.ia < 0)
            throw IntError::rangeError();
        return makeFloatValue(std::sqrt(double(a.ia)));
     case ariFloat:
        if (a.fa < 0)
            throw IntError::rangeError();
        return makeFloatValue(std::sqrt(double(a.fa)));
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FTrunc(IntValue* arg)
{
    // Truncate fractional digits
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(a.ia);
     case ariFloat:
        if (std::fabs(a.fa) > 2147483647.0)
            throw IntError::rangeError();
        return makeIntValue(int32_t(a.fa));
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FRound(IntValue* arg)
{
    // Round arithmetically
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(a.ia);
     case ariFloat:
        if (std::fabs(a.fa) > 2147483647.0)
            throw IntError::rangeError();
        if (a.fa > 0)
            a.fa += 0.5;
        else
            a.fa -= 0.5;
        return makeIntValue(int32_t(a.fa));
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FLTrim(IntValue* arg)
{
    // Trim left whitespace
    if (arg == 0)
        return 0;
    else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg))
        return makeStringValue(strLTrim(sv->getValue()));
    else
        throw IntError::typeError(IntError::tyString);
}

static IntValue*
FRTrim(IntValue* arg)
{
    // Trim right whitespace
    if (arg == 0)
        return 0;
    else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg))
        return makeStringValue(strRTrim(sv->getValue()));
    else
        throw IntError::typeError(IntError::tyString);
}

static IntValue*
FLRTrim(IntValue* arg)
{
    // Trim both whitespace
    if (arg == 0)
        return 0;
    else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg))
        return makeStringValue(strTrim(sv->getValue()));
    else
        throw IntError::typeError(IntError::tyString);
}

static IntValue*
FLength(IntValue* arg)
{
    // Get length of string
    // FIXME: PCC 1.x stringifies. Should we, too?
    if (arg == 0)
        return 0;
    else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg))
        return makeIntValue(lengthUtf8(sv->getValue()));
    else
        throw IntError::typeError(IntError::tyString);
}

static IntValue*
FVal(IntValue* arg)
{
    // Evaluate
    if (arg == 0) {
        return 0;
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(arg)) {
        string_t ssv = sv->getValue();
        if (ssv.find_first_not_of(" \t0123456789-+.") != ssv.npos) {
            // syntax error, i.e. hex
            return 0;
        } else {
            // no illegal characters, so try to convert.
            // strtod skips initial whitespace, but not trailing whitespace.
            const char* ssvp = ssv.c_str();
            char* end;
            double d = std::strtod(ssvp, &end);
            if (end == ssvp || end[std::strspn(end, " \t")] != 0)
                return 0;

            // Int or float?
            if (ssv.find('.') == ssv.npos && std::fabs(d) <= 2147483647.0)
                return makeIntValue(int32_t(d));
            else
                return makeFloatValue(d);
        }
    } else {
        throw IntError::typeError(IntError::tyString);
    }
}

static IntValue*
FTrace(IntValue* arg)
{
    // Debug log; print value and continue
    console.write(LOG_DEBUG, IntValue::toStringOf(arg, true));
    return IntValue::cloneOf(arg);
}

static IntValue*
FNot2(IntValue* arg)
{
    // Negation using binary logic: t->f, f->t, e->t
    return makeBoolValue(getBoolValue(arg) <= 0);
}

static IntValue*
FAtom(IntValue* arg)
{
    // String to atom
    if (arg == 0)
        return 0;
    return makeIntValue(atom(arg->toString(false)));
}

static IntValue*
FAtomStr(IntValue* arg)
{
    // Atom to string
    if (arg == 0)
        return 0;

    IntIntValue* iv = dynamic_cast<IntIntValue*>(arg);
    if (!iv)
        throw IntError::typeError(IntError::tyInteger);

    return makeStringValue(atomStr(iv->getValue()));
}

static IntValue*
FKeyCreate(IntValue* arg)
{
    if (arg == 0)
        return 0;

    IntStringValue* sv = dynamic_cast<IntStringValue*>(arg);
    if (!sv)
        throw IntError::typeError(IntError::tyString);

    IntKeymapRef km = IntKeymap::createKeymap(sv->getValue());
    return new IntKeymapValue(km);
}

static IntValue*
FKeyLookup(IntValue* arg)
{
    if (arg == 0)
        return 0;

    IntStringValue* sv = dynamic_cast<IntStringValue*>(arg);
    if (!sv)
        throw IntError::typeError(IntError::tyString);

    IntKeymapRef km = IntKeymap::getKeymapByName(sv->getValue());
    if (km == 0)
        throw IntError("No such keymap");
    return new IntKeymapValue(km);
}

static IntValue*
FInc(IntValue* arg)
{
    // arithmetic increment
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(a.ia + 1);
     case ariFloat:
        return makeFloatValue(a.fa + 1);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FDec(IntValue* arg)
{
    // arithmetic decrement
    ArithmeticArg a;
    switch (checkArithmetic(a, arg)) {
     case ariNull:
        return 0;
     case ariInt:
        return makeIntValue(a.ia - 1);
     case ariFloat:
        return makeFloatValue(a.fa - 1);
     default:
        throw IntError::typeError(IntError::tyNumeric);
    }
}

static IntValue*
FIsProcedure(IntValue* arg)
{
    // check for procedure
    if (!arg)
        return 0;

    if (IntCallableValue* cv = dynamic_cast<IntCallableValue*>(arg)) {
        // Callable builtin
        return makeBoolValue(cv->isProcedureCall());
    } else if (IntSubroutineValue* sv = dynamic_cast<IntSubroutineValue*>(arg)) {
        // Callable subroutine
        return makeBoolValue(sv->getBytecodeObject()->isProcedure());
    } else {
        // Variable
        return makeBoolValue(false);
    }
}

static IntValue*
FFileNr(IntValue* arg)
{
    // null
    if (!arg)
        return 0;

    // integer?
    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(arg))
        return new IntFileValue(iv->getValue());
    else if (IntFileValue* fv = dynamic_cast<IntFileValue*>(arg))
        return new IntFileValue(fv->getFileNumber());
    else
        throw IntError::typeError(IntError::tyInteger);
}

static IntValue*
FIsArray(IntValue* arg)
{
    if (!arg)
        return 0;
    else if (IntArray* a = dynamic_cast<IntArray*>(arg))
        return makeIntValue(a->getData()->getNumDimensions());
    else
        return makeIntValue(0);
}

static IntValue* (*const unary_ops[])(IntValue*) = {
    FNot,
    FBool,
    FNeg,
    FPos,
    FSin,
    FCos,
    FTan,
    FZap,
    FAbs,
    FExp,
    FLog,
    FBitNot,
    FIsEmtpy,
    FIsNum,
    FIsString,
    FAsc,
    FChr,
    FStr,
    FSqrt,
    FTrunc,
    FRound,
    FLTrim,
    FRTrim,
    FLRTrim,
    FLength,
    FVal,
    FTrace,
    FNot2,
    FAtom,
    FAtomStr,
    FKeyCreate,
    FKeyLookup,
    FInc,
    FDec,
    FIsProcedure,
    FFileNr,
    FIsArray,
};

/** Execute unary operation.
    \param op Operation, IntUnaryOperation
    \param arg User-supplied argument taken from value stack
    \return New value to push on value stack */
IntValue*
executeUnaryOperation(uint8_t op, IntValue* arg)
{
    if (op < countof(unary_ops))
        return unary_ops[op](arg);
    else
        throw IntError::internalError("invalid unary operation");
}
